#include <stdio.h>
#include <openssl/ossl_typ.h>
#include <openssl/evp.h>
#include <string.h>
#include "sm3_crypt.h"

#if 1 
struct evp_md_st {
    int type;
    int pkey_type;
    int md_size;
    unsigned long flags;
    int (*init) (EVP_MD_CTX *ctx);
    int (*update) (EVP_MD_CTX *ctx, const void *data, size_t count);
    int (*final) (EVP_MD_CTX *ctx, unsigned char *md);
    int (*copy) (EVP_MD_CTX *to, const EVP_MD_CTX *from);
    int (*cleanup) (EVP_MD_CTX *ctx);
    int block_size;
    int ctx_size;               /* how big does the ctx->md_data need to be */
    /* control function */
    int (*md_ctrl) (EVP_MD_CTX *ctx, int cmd, int p1, void *p2);
} /* EVP_MD */ ;

struct evp_md_ctx_st {
    const EVP_MD *digest;
    ENGINE *engine;             /* functional reference if 'digest' is
                                 * ENGINE-provided */
    unsigned long flags;
    void *md_data;
    /* Public key context for sign/verify */
    EVP_PKEY_CTX *pctx;
    /* Update function: usually copied from EVP_MD */
    int (*update) (EVP_MD_CTX *ctx, const void *data, size_t count);
} /* EVP_MD_CTX */ ;
#endif




static const char b64t[64] =
"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

static void do_b64_from_24bit (char **cp,
                unsigned int b2, unsigned int b1, unsigned int b0, int n)
{
        unsigned int w = (b2 << 16) | (b1 << 8) | b0;
        while (n-- > 0) {
                *(*cp)++ = b64t[w & 0x3f];
                w >>= 6;
        }
}

char *sm3_crypt(const char *key,const char *salt)
{
	EVP_MD_CTX ctx,ctx1;
	memset(&ctx,0,sizeof(EVP_MD_CTX));
	memset(&ctx1,0,sizeof(EVP_MD_CTX));
	const EVP_MD *sm3;
	const char *magic = "$8$";
	const char *name = "sm3";
        char *passwd = NULL, *p = NULL;
        const char *sp = NULL, *ep = NULL;
        unsigned char final[32] = {0};
        int sl, pl, i, j;
        unsigned long l;
        int key_len = 0;
	OpenSSL_add_all_digests();

        /* Refine the Salt first */
        sp = salt;

        /* TODO: now that we're using malloc'ed memory, get rid of the
           strange constant buffer size. */
        passwd = malloc(1024);
        if (!passwd) {
                printf("%s: Failed to malloc for passwd\n", __func__);
                return NULL;
        }
        memset(passwd, 0, 1024);
        memset(final, 0x0, sizeof final);
        key_len = (int)strlen(key);

        /* If it starts with the magic string, then skip that */
        if (!strncmp(sp, magic, strlen(magic)))
                sp += strlen(magic);

        /* It stops at the first '$', max 8 chars */
        for (ep = sp; *ep && *ep != '$' && ep < (sp + 8); ep++)
                continue;

        /* get the length of the true salt */
        sl = ep - sp;

	sm3 = EVP_get_digestbyname(name);

	EVP_MD_CTX_init(&ctx);
	EVP_DigestInit_ex(&ctx,sm3,NULL);
        /* The password first, since that is what is most unknown */
        EVP_DigestUpdate(&ctx, (unsigned const char *)key, key_len);

        /* Then our magic string */
        EVP_DigestUpdate(&ctx, (unsigned const char *)magic, strlen(magic));

        /* Then the raw salt */
        EVP_DigestUpdate(&ctx, (unsigned const char *)sp, sl);

 	/* Then just as many characters of the sm3(key, salt, key) */
	EVP_MD_CTX_init(&ctx1);
	EVP_DigestInit_ex(&ctx1,sm3,NULL);
        EVP_DigestUpdate(&ctx1, (unsigned const char *)key, key_len);
        EVP_DigestUpdate(&ctx1, (unsigned const char *)sp, sl);
        EVP_DigestUpdate(&ctx1, (unsigned const char *)key, key_len);
        EVP_DigestFinal_ex(&ctx1, final, NULL);
        for (pl = key_len; pl > 0; pl -= 32)
                EVP_DigestUpdate(&ctx, (unsigned const char *)final, pl > 32 ? 32 : pl);

        /* Don't leave anything around in vm they could use. */
        memset(final, 0, sizeof final);

        /* Then something really weird... */
        for (j = 0, i = key_len; i; i >>= 1)
                if (i & 1)
                        EVP_DigestUpdate(&ctx, (unsigned const char *)final + j, 1);
                else
                        EVP_DigestUpdate(&ctx, (unsigned const char *)key + j, 1);

        /* Now make the output string */
        strncpy(passwd, magic, strlen(magic));
        strncat(passwd, sp, sl);
        strncat(passwd, "$", 1);

        EVP_DigestFinal_ex(&ctx, final, NULL);

        /* Repeatedly run the collected hash value through SM3 to burn
         * CPU cycles.  */
        for (i = 0; i < 2017; i++) {
                sm3->init(&ctx1);
                if (i & 1)
                        EVP_DigestUpdate(&ctx1, (unsigned const char *)key, key_len);
                else
                        EVP_DigestUpdate(&ctx1, (unsigned const char *)final, 32);

                if (i % 3)
                        EVP_DigestUpdate(&ctx1, (unsigned const char *)sp, sl);

                if (i % 7)
                        EVP_DigestUpdate(&ctx1, (unsigned const char *)key, key_len);

                if (i & 1)
                        EVP_DigestUpdate(&ctx1, (unsigned const char *)final, 32);
                else
                        EVP_DigestUpdate(&ctx1, (unsigned const char *)key, key_len);
                EVP_DigestFinal_ex(&ctx1, final, NULL);
        }

        p = passwd + strlen(passwd);

        do_b64_from_24bit(&p, final[0],  final[10], final[20], 4);
        do_b64_from_24bit(&p, final[21], final[1],  final[11], 4);
        do_b64_from_24bit(&p, final[12], final[22], final[2],  4);
        do_b64_from_24bit(&p, final[3],  final[13], final[23], 4);
        do_b64_from_24bit(&p, final[24], final[4],  final[14], 4);
        do_b64_from_24bit(&p, final[15], final[25], final[5],  4);
        do_b64_from_24bit(&p, final[6],  final[16], final[26], 4);
        do_b64_from_24bit(&p, final[27], final[7],  final[17], 4);
        do_b64_from_24bit(&p, final[18], final[28], final[8],  4);
        do_b64_from_24bit(&p, final[9],  final[19], final[29], 4);
        do_b64_from_24bit(&p, 0,         final[31], final[30], 3);
        *p = '\0';

        /* Don't leave anything around in vm they could use. */
        memset(final, 0, sizeof final);

        return passwd;



}
